#ifdef HAVE_CONFIG_H
#include <libucdaemon_config.h>
#endif

#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <syslog.h>
#include <unistd.h>

#include <libucdaemon.h>

#define UCD_POLL_EVENTS	128

struct ucd_poll_fd
{
	int fd;
	int flags;
	ucd_fd_handler_t cb;
	intptr_t cbdata;
	struct ucd_poll_fd *next;
	struct ucd_poll_fd *prev;
};

struct ucd_poll
{
	int epfd;
	struct ucd_poll_fd *list;
};

static int
ucd_poll_flags_to_ep_events(int flags)
{
	int events = EPOLLHUP;
	if ((flags & UCD_POLL_READ) != 0) {
		events |= EPOLLIN | EPOLLPRI;
	}
	if ((flags & UCD_POLL_WRITE) != 0) {
		events |= EPOLLOUT;
	}
	if ((flags & UCD_POLL_ONESHOT) != 0) {
		events |= EPOLLONESHOT;
	}
	return events;
}

static int
ucd_poll_flags_from_ep_events(int events)
{
	if ((events & EPOLLOUT) != 0) {
		return UCD_POLL_WRITE;
	}
	if ((events & (EPOLLIN | EPOLLPRI)) != 0) {
		return UCD_POLL_READ;
	}
	return 0;
}

ucd_poll_t
ucd_poll_open(int size)
{
	struct ucd_poll *pp;

	pp = malloc(sizeof(ucd_poll_t));
	if (pp == NULL) {
		ucd_error("out of memory");
	}
	memset(pp, 0, sizeof(ucd_poll_t));

	pp->epfd = epoll_create(size);
	if (pp->epfd < 0) {
		ucd_error("Failed to open epoll: %s", strerror(errno));
	}

	return (ucd_poll_t) pp;
}

int
ucd_poll_close(ucd_poll_t pd)
{
	struct ucd_poll *pp;
	struct ucd_poll_fd *pfdp;

	pp = (struct ucd_poll *) pd;
	if (pp == NULL) {
		return -1;
	}

	while (pp->list != NULL) {
		pfdp = pp->list;
		pp->list = pfdp->next;
		free(pfdp);
	}

	close(pp->epfd);
	free(pp);

	return 0;
}

ucd_poll_fd_t
ucd_poll_add_fd(ucd_poll_t pd, int fd, int flags, ucd_fd_handler_t cb, intptr_t cbdata)
{
	struct ucd_poll *pp;
	struct ucd_poll_fd *pfdp;
	struct epoll_event event;

	pp = (struct ucd_poll *) pd;
	if (pp == NULL || cb == NULL) {
		return (ucd_poll_fd_t) NULL;
	}

	pfdp = malloc(sizeof(ucd_poll_fd_t));
	if (pfdp == NULL) {
		ucd_error("out of memory");
	}
	memset(pfdp, 0, sizeof(ucd_poll_fd_t));

	pfdp->fd = fd;
	pfdp->flags = flags;
	pfdp->cb = cb;
	pfdp->cbdata = cbdata;

	event.events = ucd_poll_flags_to_ep_events(flags);
	event.data.ptr = pfdp;

	if (epoll_ctl(pp->epfd, EPOLL_CTL_ADD, fd, &event) < 0) {
		free(pfdp);
		ucd_log(LOG_ERR, "Failed to add file descriptor with epoll_ctl: %s", strerror(errno));
		return (ucd_poll_fd_t) NULL;
	}

	pfdp->next = pp->list;
	pp->list = pfdp;

	return (ucd_poll_fd_t) pfdp;
}

int
ucd_poll_modify_fd(ucd_poll_t pd, ucd_poll_fd_t pfd, int flags)
{
	struct ucd_poll *pp;
	struct ucd_poll_fd *pfdp;
	struct epoll_event event;

	pp = (struct ucd_poll *) pd;
	pfdp = (struct ucd_poll_fd *) pfd;
	if (pp == NULL || pfdp == NULL) {
		return -1;
	}

	event.events = ucd_poll_flags_to_ep_events(flags);
	event.data.ptr = pfdp;

	if (epoll_ctl(pp->epfd, EPOLL_CTL_MOD, pfdp->fd, &event) < 0) {
		ucd_log(LOG_ERR, "Failed to modify file descriptor with epoll_ctl: %s", strerror(errno));
		return -1;
	}

	return 0;
}

int
ucd_poll_delete_fd(ucd_poll_t pd, ucd_poll_fd_t pfd)
{
	struct ucd_poll *pp;
	struct ucd_poll_fd *pfdp;
	struct epoll_event event;

	pp = (struct ucd_poll *) pd;
	pfdp = (struct ucd_poll_fd *) pfd;
	if (pp == NULL || pfdp == NULL) {
		return -1;
	}

	if (epoll_ctl(pp->epfd, EPOLL_CTL_DEL, pfdp->fd, &event) < 0) {
		ucd_log(LOG_ERR, "Failed to delete file descriptor with epoll_ctl: %s", strerror(errno));
		return -1;
	}

	if (pfdp->next != NULL) {
		pfdp->next->prev = pfdp->prev;
	}
	if (pfdp->prev != NULL) {
		pfdp->prev->next = pfdp->next;
	} else {
		pp->list = pfdp->next;
	}
	free(pfdp);

	return 0;
}

int
ucd_poll_dispatch(ucd_poll_t pd, int timeout)
{
	struct ucd_poll *pp;
	struct ucd_poll_fd *pfdp;
	struct epoll_event events[UCD_POLL_EVENTS];
	int nfd, i, fd;
	ucd_fd_handler_t cb;
	intptr_t cbdata;

	pp = (struct ucd_poll *) pd;
	if (pp == NULL) {
		return -1;
	}

	nfd = epoll_wait(pp->epfd, events, UCD_POLL_EVENTS, timeout);
	if (nfd < 0) {
		ucd_log(LOG_ERR, "Failed epoll_wait: %s", strerror(errno));
		return -1;
	}

	for (i = 0; i < nfd; i++) {
		pfdp = events[i].data.ptr;

		fd = pfdp->fd;
		cb = pfdp->cb;
		cbdata = pfdp->cbdata;

		if ((pfdp->flags & UCD_POLL_ONESHOT) != 0) {
			ucd_poll_delete_fd(pd, (ucd_poll_fd_t) pfdp);
		}

		(*cb)(fd, ucd_poll_flags_from_ep_events(events[i].events), cbdata);
	}

	return nfd;
}
